import unittest

from PySide6 import QtCore, QtWidgets

from qtbricks import widgets


class TestQHLine(unittest.TestCase):
    def setUp(self):
        self.app = (
            QtWidgets.QApplication.instance() or QtWidgets.QApplication()
        )
        self.widget = widgets.QHLine()
        self.addCleanup(self.release_qt_resources)

    def release_qt_resources(self):
        self.widget.deleteLater()
        self.app.sendPostedEvents(event_type=QtCore.QEvent.DeferredDelete)
        self.app.processEvents()

    def test_instantiate_class(self):
        pass


class TestQVLine(unittest.TestCase):
    def setUp(self):
        self.app = (
            QtWidgets.QApplication.instance() or QtWidgets.QApplication()
        )
        self.widget = widgets.QVLine()
        self.addCleanup(self.release_qt_resources)

    def release_qt_resources(self):
        self.widget.deleteLater()
        self.app.sendPostedEvents(event_type=QtCore.QEvent.DeferredDelete)
        self.app.processEvents()

    def test_instantiate_class(self):
        pass



import unittest

from PySide6 import QtCore, QtWidgets, QtTest

from qtbricks import aboutdialog, testing


class TestDatasetDisplayWidget(unittest.TestCase):
    def setUp(self):
        self.app = (
            QtWidgets.QApplication.instance() or QtWidgets.QApplication()
        )
        self.package_name = "qtbricks"
        self.widget = aboutdialog.AboutDialog(package_name=self.package_name)
        self.addCleanup(self.release_qt_resources)

    def release_qt_resources(self):
        self.widget.deleteLater()
        self.app.sendPostedEvents(event_type=QtCore.QEvent.DeferredDelete)
        self.app.processEvents()

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_window_title(self):
        self.assertEqual(
            f"About {self.package_name}", self.widget.windowTitle()
        )

    def test_has_close_button(self):
        close_button = self.widget._button_box.button(
            QtWidgets.QDialogButtonBox.Close
        )
        self.assertTrue(close_button)


class TestDatasetDisplayWidgetSignals(testing.TestCaseUsingQSignals):
    def setUp(self):
        super().setUp()
        self.package_name = "qtbricks"
        self.widget = aboutdialog.AboutDialog(package_name=self.package_name)

    def test_pressing_close_button_closes_dialog(self):
        close_button = self.widget._button_box.button(
            QtWidgets.QDialogButtonBox.Close
        )
        with self.assertSignalReceived(self.widget.rejected, None):
            QtTest.QTest.mouseClick(
                close_button,
                QtCore.Qt.MouseButton.LeftButton,
            )



import os
import unittest

from PySide6 import QtCore, QtWidgets

from qtbricks import utils


class TestImagePath(unittest.TestCase):
    def setUp(self):
        self.basedir = os.path.join(
            os.path.split(os.path.dirname(__file__))[0], "qtbricks"
        )

    def test_image_path_with_defaults(self):
        image_name = "foo.svg"
        expected = os.path.join(self.basedir, "images", image_name)
        actual = utils.image_path(name=image_name)
        self.assertEqual(expected, actual)

    def test_image_path_with_image_dir(self):
        image_name = "foo.svg"
        image_path = "icons"
        expected = os.path.join(self.basedir, image_path, image_name)
        actual = utils.image_path(name=image_name, image_dir=image_path)
        self.assertEqual(expected, actual)

    def test_image_path_with_base_dir(self):
        image_name = "foo.svg"
        base_dir = os.path.dirname(__file__)
        expected = os.path.join(base_dir, "images", image_name)
        actual = utils.image_path(name=image_name, base_dir=base_dir)
        self.assertEqual(expected, actual)


class TestCreateButton(unittest.TestCase):
    def setUp(self):
        self.app = (
            QtWidgets.QApplication.instance() or QtWidgets.QApplication()
        )
        self.addCleanup(self.release_qt_resources)

    def release_qt_resources(self):
        self.app.sendPostedEvents(event_type=QtCore.QEvent.DeferredDelete)
        self.app.processEvents()

    def test_create_button_returns_button(self):
        button = utils.create_button()
        self.assertIsInstance(button, QtWidgets.QPushButton)


class TestIntValidator(unittest.TestCase):
    def setUp(self):
        self.validator = utils.IntValidator()

    def test_instantiate_class(self):
        pass

    def test_fixup_with_value_exceeding_top_returns_top(self):
        self.validator.setTop(42)
        self.assertEqual(
            self.validator.fixup(str(42 + 5)),
            str(self.validator.top()),
        )

    def test_fixup_with_value_exceeding_bottom_returns_bottom(self):
        self.validator.setBottom(42)
        self.assertEqual(
            self.validator.fixup(str(42 - 5)),
            str(self.validator.bottom()),
        )



"""
Main GUI window.

There is usually one main window of a GUI, and this window consists of all
the default parts provided by the Qt framework, namely menu bar and toolbars
at the top, status bar at the bottom, and areas for dockable windows/widgets
on all four sides of the central widget.

While the actual contents of both, the central widget and the dock areas
and the other components are more or less unique for your application,
a main window of a GUI comes with a lot of features and basic
functionality that are rather general. Hence you can greatly reduce the
amount of code to write when inheriting from the :class:`MainWindow` class
implemented here.

.. note::

    Some ideas are borrowed and adapted from Mark Summerfield and his
    excellent book "Rapid GUI Programming with Python and Qt" (Prentice
    Hall, Upper Saddle River, 2008). While written for Python 2 and PyQt4,
    both the concepts and the thorough introduction to GUI programming are
    probably still the best that is available for creating GUIs with
    Python and Qt.


How to call and where to instantiate
====================================

Rather than creating an instance of :class:`MainWindow` yourself, you will
usually call the :func:`app.main` function of the :mod:`app` module,
or simply call the GUI by means of the respective gui_scripts entry point
defined in ``setup.py``.


.. note::

    Of course, you will usually not use the :class:`MainWindow` as
    implemented here, as it lacks all contents, but rather create your own
    ``MainWindow`` class inheriting from :class:`MainWindow` and setting
    the central widgets and everything else as appropriate.


An example of an :mod:`app` module is shown below. Note the fictitious
name ``mypackage`` that you would need to replace with the actual name of
your package:

.. code-block::

    import sys

    from PySide6.QtWidgets import QApplication
    from PySide6.QtGui import QIcon

    from qtbricks import utils

    from mypackage import mainwindow

    def main():
        app = QApplication(sys.argv)
        app.setOrganizationName("demoapp")
        app.setOrganizationDomain("example.org")
        app.setApplicationName("Demo application")
        app.setWindowIcon(QIcon(utils.image_path("icon.svg")))

        window = mainwindow.MainWindow()
        window.show()
        app.exec()


    if __name__ == "__main__":
        main()


Note that the setting of the organisation name
(``app.setOrganizationName()``) is crucial for storing settings, as this
determines the name of the directory the settings are stored under. The
exact location of the configuration depends on your operating system.
Furthermore, setting the application name (``app.setApplicationName()``)
is relevant for programmatically obtaining the application name, *e.g.*,
for window titles and alike. As you can see from the above code,
Qt is capable of handling SVG files directly, very convenient for icons
and logos of your application.

The corresponding section in the ``setup.py`` file with the gui entrypoint
may look similar to the following:

.. code-block::

    setuptools.setup(
        # ...
        entry_points={
            "gui_scripts": [
                "demoapp = demoapp.gui.app:main"
            ]
        },
        # ...


In this particular case, both, your package and the respective GUI
entrypoint, *i.e.* the callable from the terminal, are named "demoapp". Of
course, you will need to change this to some sensible name for your
package/GUI application.


Some notes for developers
=========================

GUI programming can be quite complex, and the :class:`MainWindow` class
provides only a small set of (sensible) defaults for the most common tasks.
However, to help with creating GUIs and to keep the code as readable as
possible, some general advice and an overview of the functionality
implemented are given below.


First step: creating your own mainwindow module
-----------------------------------------------

The first step is always to create your own ``mainwindow`` module,
presumably in the ``gui`` subpackage of your package. A rather minimal
structure of this module is shown below:


.. code-block::

    from PySide6 import QtWidgets

    import qtbricks.mainwindow

    import mypackage.gui.model as model


    class MainWindow(qtbricks.mainwindow.MainWindow):

        def __init__(self):
            # Customise your main window, setting at least a central widget.
            # Try to keep methods as short and concise as possible.


A few comments, beyond the obvious:

* You will want/need to import further modules containing the definition
  of the widget(s) used as central widget and possibly in dockable areas.

* You should always follow the model-view paradigm for larger
  applications, and come up with a model of your application containing
  the business logic. This is reflected in the import of a model module of
  your fictitious ``mypackage`` package.


Saving and restoring GUI settings
---------------------------------

By default, window geometry and state will be saved on close and restored on
startup. This creates a file, typically in the user's home directory,
and depending on the respective platform. Directory and file name depend on
the settings of organisation and application name on the application level.
For details, see the :func:`app.main` function in the :mod:`app` module.

To this end, a private method :meth:`MainWindow._restore_settings` gets
called from the constructor to restore the settings, and the
:meth.`MainWindow.closeEvent` method is overridden to take care of saving
the current settings on closing the GUI.


Adding menus
------------

Menus (and toolbars) are essential aspects of nearly every GUI, particular
the main window. In its default configuration, three menus are created:
"File", "View", and "Help". The "File" menu is populated with a "Quit"
action and associated keyboard shortcut "Ctrl + Q", the "Help" menu with an
"About" action and associated keyboard shortcut "F1". The "View" menu is a
reminder that in case of dockable windows, you need a way to restore windows
closed by the user.

While the :meth:`MainWindow._create_menus` method gets called from the
:meth:`MainWindow._setup_ui` method on startup, for the sake of code
readability it is best to separate creating each menu into its own (private)
method that gets called in turn from within the
:meth:`MainWindow._create_menus` method.

Each (active) entry in a menu is an action in Qt speak, and to help with
adding those actions, there is a (private) method
:meth:`MainWindow._create_action` taking a series of parameters. Perhaps it
is best explained with an example:


.. code-block::

    file_quit_action = self._create_action(
        "&Quit",
        self.close,
        QtGui.QKeySequence.Quit or "Ctrl+Q",
        "file_quit",
        "Close the application"
    )


Here, we define the action for quitting the main window and hence the entire
GUI application. The first argument is the text appearing in the menu,
including the keyboard accelerator character. The second argument is the
actual method being called, followed by the keyboard shortcut, a name for an
icon file (entirely optional), and finally an explaining text used as
tooltip in case of a toolbar and shown in the status bar in case of a menu.

Qt actions are defined once and can be used in different contexts: You can
add them to a menu, to a context menu, and to a toolbar. Depending on your
way of organising your code, you will need to either make the actions
(private) attributes of the :class:`MainWindow` class or define menus,
context menus and toolbars all within one method.

A further convenience method helping to add actions to menus is
:meth:`MainWindow._add_actions`. Here, you can not only provide a list of
actions, but "None" as well that will be converted to separators.


Adding dockable windows
-----------------------

Generally, dockable windows need to inherit from
:class:`PySide6.QtWidgets.QDockWidget`. A prototypical class for a dockable
window is shown below. Adapt to your own needs. Of course, a dockable window
can (and usually will) contain arbitrarily complex widgets.


.. code-block::

    class GeneralDockWindow(QtWidgets.QDockWidget):

        def __init__(self, title="General Dock Window"):
            super().__init__()
            self.setObjectName("GeneralDockWindow")
            self.setWindowTitle(title)
            self.list_widget = QtWidgets.QListWidget()
            self.setWidget(self.list_widget)


Note that a very similar implementation is provided in the
:class:`GeneralDockWindow` class. See its documentation for further details.

The possible docking areas the window can be attached to can be restricted
as well:


.. code-block::

    self.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea)


To add such a dockable window to your main window, extend the
:meth:`MainWindow._create_dock_windows` method in a way similar to the
following:


.. code-block::

    def _create_dock_windows(self):
        general_dock_window_1 = GeneralDockWindow("general 1")
        self.addDockWidget(Qt.LeftDockWidgetArea, general_dock_window_1)
        self._view_menu.addAction(general_dock_window_1.toggleViewAction())


Important here is to add the action to the View menu in order to be able to
restore a closed dockable window. To make things easier, the same can be
achieved using the convenience function :meth:`MainWindow._add_dock_window`:


.. code-block::

    dock_window = GeneralDockWindow()
    self._add_dock_window(
        dock_window,
        "general 1",
        Qt.LeftDockWidgetArea
    )


In case you want to layout two or more dock widgets as tabbed windows,
use the same method, but assign the output to a variable of your choosing,
and afterwards programmatically "tabify" the respective docked windows:


.. code-block::

    dock1 = self._add_dock_window(...)
    dock2 = self._add_dock_window(...)
    self.tabifyDockWidget(dock1, dock2)


Just make sure that both windows you want to appear in tabs are set to the
same docking area. Otherwise, the first one will "win" and drag the second
one to the same docking area.


Preventing loss of work on closing the GUI
------------------------------------------

As soon as the user can change contents from within the GUI, it may no
longer be safe to simply close the window and hence quit the entire
application. The same is true for changes that need to be applied or
discarded before switching to some other dataset or else. To detect (and
handle) the situation when changes are present that have not yet been saved,
reimplement the method :meth:`MainWindow._ok_to_continue` according to your
needs. Furthermore, this method allows for convenient and readable code
similar to


.. code-block::

    if self._ok_to_continue():
        ...


For an actual example, you may have a look at the
:meth:`MainWindow.closeEvent` method.


Module documentation
====================

"""

from PySide6 import QtGui, QtWidgets
from PySide6.QtCore import SIGNAL, QSettings, QByteArray, QSize, Qt

from qtbricks import utils, aboutdialog


class MainWindow(QtWidgets.QMainWindow):
    """
    Main GUI window of the application.

    There is usually one main window of a GUI, and this window consists of
    all the default parts provided by the Qt framework, namely menu bar and
    toolbars at the top, status bar at the bottom, and areas for dockable
    windows/widgets on all four sides of the central widget.

    While any more complex dialogs and widgets are usually designed using
    the QtDesigner tool, the main window, consisting typically of one
    central widget, is laid out programmatically.

    Rather than creating an instance of :class:`MainWindow` yourself,
    you will usually call the :func:`app.main` function of the :mod:`app`
    module, or simply call the GUI by means of the respective gui_scripts entry
    point defined in ``setup.py``.

    By default, window geometry and state will be saved on close and
    restored on startup. This creates a file, typically in the user's home
    directory, and depending on the respective platform. Directory and file
    name depend on the settings of organisation and application name on the
    application level. For details, see the :func:`app.main` function in the
    :mod:`app` module.

    Attributes
    ----------
    package_name : :class:`str`
        Name of the package the mainwindow belongs to.

        This information is required, *i.a.*, for the "Help About" window.

    logo : :class:`str`
        Path to logo image file

        The logo is used, *i.a.*, for the Help About window.

    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.package_name = "qtbricks"
        self.logo = utils.image_path("icon.svg")

        self._view_menu = None

        self._setup_ui()
        self._restore_settings()

    def _restore_settings(self):
        settings = QSettings()
        self.restoreGeometry(
            settings.value("MainWindow/Geometry", QByteArray())
        )
        self.restoreState(settings.value("MainWindow/State", QByteArray()))

    def _setup_ui(self):
        """
        Create the elements of the main window.

        The elements created are:

        #. The central widget
        #. The status bar
        #. The menu bar
        #. The dockable windows

        All these tasks are delegated to (non-public) methods. The toolbar
        will usually be created from within the menus, as the actions will
        be defined therein.

        In its default configuration, three menus are created: "File",
        "View", and "Help". The "File" menu is populated with a "Quit"
        action and associated keyboard shortcut "Ctrl + Q", the "Help" menu
        with an "About" action and associated keyboard shortcut "F1". The
        "View" menu is a reminder that in case of dockable windows, you need
        a way to restore windows closed by the user.

        Additionally, the minimum size and title of the main GUI window is
        set. The window title is identical to the application name set in
        the :func:`app.main` function in the :mod:`app` module.

        """
        self._create_central_widget()
        self._create_status_bar()
        self._create_menus()
        self._create_dock_windows()

        self.setMinimumSize(QSize(640, 480))
        self.setWindowTitle(QtWidgets.QApplication.applicationName())

    def _create_central_widget(self):
        pass

    def _create_status_bar(self):
        status = self.statusBar()
        status.setSizeGripEnabled(False)
        status.showMessage("Ready", 5000)

    def _create_menus(self):
        self._create_file_menu()
        self._create_view_menu()
        self._create_help_menu()

    def _create_dock_windows(self):
        pass

    def _create_file_menu(self):
        # noinspection PyUnresolvedReferences
        file_quit_action = self._create_action(
            "&Quit",
            self.close,
            QtGui.QKeySequence.Quit or "Ctrl+Q",
            "file_quit",
            "Close the application",
        )
        file_menu = self.menuBar().addMenu("&File")
        self._add_actions(file_menu, (file_quit_action,))

    def _create_view_menu(self):
        self._view_menu = self.menuBar().addMenu("&View")

    def _create_help_menu(self):
        help_about_action = self._create_action(
            "&About", self.help_about, "F1", "", "About the application"
        )
        help_menu = self.menuBar().addMenu("&Help")
        self._add_actions(help_menu, (help_about_action,))

    # pylint: disable=too-many-arguments
    def _create_action(
        self,
        text,
        slot=None,
        shortcut=None,
        icon=None,
        tip=None,
        checkable=False,
        signal="triggered()",
    ):
        action = QtGui.QAction(text, self)
        if icon:
            action.setIcon(QtGui.QIcon(utils.image_path(f"{icon}.png")))
        if shortcut:
            action.setShortcut(shortcut)
        if tip:
            action.setToolTip(tip)
            action.setStatusTip(tip)
        if slot:
            # noinspection PyTypeChecker
            self.connect(action, SIGNAL(signal), slot)
        if checkable:
            action.setCheckable(True)
        return action

    @staticmethod
    def _add_actions(target, actions):
        for action in actions:
            if action is None:
                target.addSeparator()
            else:
                target.addAction(action)

    def _add_dock_window(
        self, dock_window=None, title="", area=Qt.RightDockWidgetArea
    ):
        if title:
            dock_window.setTitle(title)
        self.addDockWidget(area, dock_window)
        self._view_menu.addAction(dock_window.toggleViewAction())
        return dock_window

    def help_about(self):
        """
        Show a dialog with basic information about the application.

        Presenting the user with a concise summary of what the entire
        application is about, perhaps together with the author names,
        a license information, and some very basic system settings,
        is a sensible thing to do. Typically, this can be found in the "Help
        -> About" menu.
        """
        dialog = aboutdialog.AboutDialog(
            parent=self, package_name=self.package_name, logo=self.logo
        )
        dialog.exec()

    def closeEvent(self, event):  # noqa N802
        """
        Actions performed when attempting to close the window.

        By default, both geometry and state of the main window are saved to
        the settings file, to be restored on startup, using the standard Qt
        machinery for this purpose.
        """
        if self._ok_to_continue():
            settings = QSettings()
            settings.setValue("MainWindow/Geometry", self.saveGeometry())
            settings.setValue("MainWindow/State", self.saveState())
        else:
            event.ignore()

    # noinspection PyMethodMayBeStatic
    def _ok_to_continue(self):
        """
        Helper method determining whether it is safe to continue.

        Often, certain actions shall not be performed if the application is in
        an unsaved state (*e.g.*, closing the application). Therefore,
        implement this method according to your needs. In its default
        implementation, it always returns "True".

        Returns
        -------
        status : :class:`bool`
            Whether it is safe to continue

        """
        return True


class GeneralDockWindow(QtWidgets.QDockWidget):
    """
    Convenience class for dockable windows.

    This is a thin wrapper for widgets that should appear as dockable
    windows in a main application window. Note that this dock window is not
    restricted with respect to the dockable areas it can be positioned in.
    You may, however, restrict this afterwards. See the examples section for
    details.

    Parameters
    ----------
    title : :class:`str`
        Window title

        The window title should be comprehensive, as it is used both,
        as window title in the dock and in the view menu of the main window.

    widget : :class:`PySide6.QtWidgets.QWidget`
        Widget to be set as (central) widget of the dockable window


    Examples
    --------
    To add a dockable window to your main application window, you first need
    to create a dockable window containing the (complex) widget of your
    choosing, and afterwards add it to your main application window.

    If you subclass :class:`MainWindow` for your main application window,
    you can make use of a series of convenience methods provided. Dockable
    windows should be defined within the method
    :meth:`MainWindow._create_dock_windows`. You can add the dockable window
    using the method :meth:`MainWindow._add_dock_window`.

    .. code-block::

        def _create_dock_windows(self):
            dock_window = qtbricks.mainwindow.GeneralDockWindow(
                title="My fancy dockable window",
                widget=QtWidgets.QListWidget(),
                object_name="MyDockWindow"
            )
            self._add_dock_window(dock_window=dock_window)

    This will add a dock window to the main application window. Make sure to
    set the object name, as otherwise, Qt will produce warnings if you try
    to save the application state.

    If you would like to restrict the possible docking areas the dock window
    can be added to, you may set the respective property after initialising
    the object:

    .. code-block::

        dock_window = qtbricks.mainwindow.GeneralDockWindow(
            title="My fancy dockable window",
            widget=QtWidgets.QListWidget(),
            object_name="MyDockWindow"
        )
        dock_window.setAllowedAreas(
            Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea
        )

    Note that you need to import ``PySide6.QtCore.Qt`` for this to work.

    """

    def __init__(
        self, title="General dock window", object_name="", widget=None
    ):
        super().__init__()
        self.setWindowTitle(title)
        self.setObjectName(object_name)
        # noinspection PyTypeChecker
        self.setWidget(widget)


def _main():
    """
    Entry point for the GUI application.

    This function serves as main entry point to the GUI application and gets
    added as "gui_script" entry point. Additionally, the essential
    aspects of the (Qt) application are set that are relevant for saving and
    restoring settings, as well as the window icon.
    """
    app = QtWidgets.QApplication(sys.argv)
    app.setOrganizationName("qtbricks")
    app.setOrganizationDomain("example.org")
    app.setApplicationName("Demo application")
    app.setWindowIcon(QtGui.QIcon(utils.image_path("icon.svg")))

    window = MainWindow()
    window.show()
    app.exec()


if __name__ == "__main__":
    import sys

    _main()



"""
Testing-related helper functionality of the qtbricks package.

GUIs should be tested, as every code. However, testing GUIs and widgets is
somewhat more complicated and often involves "faking" interaction with the
different widgets.

In case of Qt, even setting up a test in a way that allows for testing any
GUI-related stuff needs a bit of additional knowledge.

This module contains a growing series of helper classes and functions
designed to make your test code simpler and easier to read.
"""

import unittest

from PySide6 import QtCore, QtTest, QtWidgets


def qtest_enter_text(widget=None, text=""):
    """
    Convenience function to enter text in a QLineEdit widget.

    When testing GUI widgets, you sometimes need to enter text into a line
    edit widget. However, entering/replacing text in such a widget is a
    three-step process:

    #. Clear the current text.
    #. Enter the text.
    #. Press the return key to fire the correct signals.

    For convenience, this function takes care of all three steps.

    Parameters
    ----------
    widget : :class:`PySide6.QtWidgets.QLineEdit`
        Widget to enter text into.

    text : :class:`str`
        Text to enter into the widget.

    """
    widget.clear()
    QtTest.QTest.keyClicks(widget, text)
    QtTest.QTest.keyPress(widget, QtCore.Qt.Key.Key_Return)


class SignalReceiver:
    """
    Context manager class detecting whether a QSignal has been emitted.

    Adapted from https://stackoverflow.com/a/48128768
    """

    def __init__(self, test, signal, *args):
        self.test = test
        self.signal = signal
        self.called = False
        self.expected_args = args
        self.actual_args = None

    def slot(self, *args):
        """
        Slot connected to the signal to test for.

        If the signal is emitted, the attribute :attr:`called` is set to True.

        Parameters
        ----------
        args
            Parameters passed by the emitted signal.

        """
        self.actual_args = args
        self.called = True

    def __enter__(self):
        """Entrance of the context manager."""
        self.signal.connect(self.slot)

    def __exit__(self, exception, msg, traceback):
        """
        Exit of the context manager.

        Parameters
        ----------
        exception
            Exception that may have been raised.

        msg : :class:`str`
            Message to pass to the exception

        traceback
            Call trace leading to the exception

        """
        if exception:
            raise exception(msg)
        self.test.assertTrue(self.called, "Signal not called!")
        if self.actual_args:
            self.test.assertEqual(
                self.expected_args,
                self.actual_args,
                f"""Signal arguments don't match!
                actual:   {self.actual_args}
                expected: {self.expected_args}""",
            )


class SignalNotReceiver:
    """
    Context manager class detecting whether a QSignal has not been emitted.

    Adapted from https://stackoverflow.com/a/48128768
    """

    def __init__(self, test, signal):
        self.test = test
        self.signal = signal
        self.called = False

    def slot(self):
        """
        Slot connected to the signal to test for.

        If the signal is emitted, the attribute :attr:`called` is set to True.
        """
        self.called = True

    def __enter__(self):
        """Entrance of the context manager."""
        self.signal.connect(self.slot)

    def __exit__(self, exception, msg, traceback):
        """
        Exit of the context manager.

        Parameters
        ----------
        exception
            Exception that may have been raised.

        msg : :class:`str`
            Message to pass to the exception

        traceback
            Call trace leading to the exception

        """
        if exception:
            raise exception(msg)
        self.test.assertFalse(self.called, "Signal called!")


class TestCaseUsingQSignals(unittest.TestCase):
    """
    Test class for testing whether a QSignal has been emitted.

    Adapted from https://stackoverflow.com/a/48128768
    """

    def setUp(self):
        """Create the QApplication instance"""
        _instance = QtWidgets.QApplication.instance()
        if not _instance:
            _instance = QtWidgets.QApplication([])
        self.app = _instance

    def tearDown(self):
        """Delete the reference owned by self"""
        del self.app

    def assertSignalReceived(self, signal, args):  # noqa N802
        """
        Check whether signal has been received.

        Parameters
        ----------
        signal
            Signal to check for

        args
            Parameters passed by the emitted signal.

        Returns
        -------
        answer : :class:`bool`
            Result of the test

        """
        return SignalReceiver(self, signal, args)

    def assertSignalNotReceived(self, signal):  # noqa N802
        """
        Check whether signal has not been received.

        Parameters
        ----------
        signal
            Signal to check for

        Returns
        -------
        answer : :class:`bool`
            Result of the test

        """
        return SignalNotReceiver(self, signal)



"""
General utils of the qtbricks package.

Mostly small functions that are more general and get used in the other
modules. Some of these functions may be useful for packages importing
qtbricks, hence generally the individual functions are designed with
reusability beyond the qtbricks package in mind.
"""

import os

from PySide6 import QtWidgets, QtGui


def image_path(name="", image_dir="images", base_dir=""):
    """
    Return full path to a given image.

    Images, such as icons, are used in several places in GUIs, but are
    usually stored in a separate directory. Hence, a generic function
    returning the full path is both, convenient and modular.

    If the internal organisation of images changes, only this code needs to
    be adapted.

    For use with own packages, you may want to set the parameter
    ``base_dir`` accordingly.

    Parameters
    ----------
    name : :class:`str`
        Name of the icon

    image_dir : :class:`str`
        Directory containing the icons

        Default: "images"

    base_dir : :class:`str`
        Directory used as base for the icons directory

        Useful in cases where the function should be used outside this
        package, *i.e.* with a different base directory for the images.

        Default: ``os.path.dirname(__file__)``

    Returns
    -------
    path : :class:`str`
        Full path to the icon

    """
    base_dir = base_dir or os.path.dirname(__file__)
    path = os.path.join(base_dir, image_dir, name)
    return path


# pylint: disable=too-many-arguments
def create_button(
    text="", slot=None, shortcut="", icon="", checkable=False, tooltip=""
):
    """
    Conveniently create a button.

    Parameters
    ----------
    text : :class:`str`
        The text appearing on the button.

        In case you provide both, icon (see below) and text, both are
        displayed. If you intend to only present an icon, set the text to an
        empty string.

        Default: empty

    slot : :py:obj:`function <types.FunctionType>`
        The slot to connect to the button.

    shortcut : :class:`str`
        The keyboard shortcut used for the action connected to the button.

        Default: empty

    icon : :class:`str`
        The name of the icon file to be used.

        If empty, no icon will be displayed.

        For enhanced modularity, do *not* provide paths, but only icon file
        names, as the full path will get looked up using the
        :func:`icon_path` function.

    checkable : :class:`bool`
        Whether the button is checkable, *i.e.*, displays its current state.

    tooltip : class:`str`
        Tooltip to be displayed for the button.

        Sensible tooltips go a long way towards a user-friendly GUI.

    Returns
    -------
    button : :class:`QtWidgets.QPushButton`

    """
    button = QtWidgets.QPushButton(text)
    if icon:
        button.setIcon(QtGui.QIcon(image_path(icon)))
    if shortcut:
        button.setShortcut(shortcut)
    if tooltip:
        if shortcut:
            tooltip = "\n".join([tooltip, f"Keyboard shortcut: {shortcut}"])
        button.setToolTip(tooltip)
    if slot:
        if checkable:
            button.toggled.connect(slot)
        else:
            button.pressed.connect(slot)
    if checkable:
        button.setCheckable(True)
    return button


def make_buttons_in_group_uncheckable(buttongroup):
    """
    Allow all buttons in exclusive button group to be unchecked.

    By default, an exclusive button group does not allow to uncheck all
    buttons by clicking the selected button, but only to select an
    alternative button, once one button has been clicked.

    Parameters
    ----------
    buttongroup : :class:`QtWidgets.QButtonGroup`
        Button group to operate on

    """
    buttongroup.buttonPressed.connect(
        lambda button: button.group().setExclusive(not button.isChecked())
    )
    buttongroup.buttonClicked.connect(
        lambda button: button.group().setExclusive(True)
    )


class IntValidator(QtGui.QIntValidator):
    """
    Integer validator actually fixing input that is beyond the boundaries.

    The standard Qt integer validator (actually the base class of this class)
    has an empty `fixup` method. Here, the fixup sets text exceeding the set
    boundaries to the respective boundary.

    Examples
    --------
    A typical use case of the :class:`IntValidator` class are
    :class:`PySide6.QtWidgets.QLineEdit` widgets:

    .. code-block::

        line_edit = QtWidgets.QLineEdit()
        validator = IntValidator(0, 42)
        line_edit.setValidator(validator)

    In this case, if the user enters values smaller than 0 or larger than 42,
    the actual value of the line edit will be set to the respective boundary.

    """

    def fixup(self, value):
        """
        Attempt to change input to be valid according to the validator rules.

        In case the value exceeds the lower or upper boundary defined by
        :attr:`bottom` or :attr:`top`, the respective boundary is returned.

        Parameters
        ----------
        value : :class:`str`
            Value to be fixed.

        Returns
        -------
        value : :class:`str`
            Value that has been fixed.

        """
        if int(value) > self.top():
            value = str(self.top())
        if int(value) < self.bottom():
            value = str(self.bottom())
        return value



"""
Plot widget for including Matplotlib plots in a GUI.

Graphical representations of data are an often-found use case in scientific
GUIs. The de-facto framework for such plotting in Python in Matplotlib,
hence this widget to include a Matplotlib figure into a Qt GUI.


General characteristics and intended purpose
============================================

Ideally, the programmer dealing with the actual plotting tasks does not need
to care in any way where the plot ends up, hence the respective figure and
axes objects are exposed as public attributes of the widget.

Plot windows exposed to users for interaction should typically provide a
numer of functions, such as zooming and panning. While the default
Matplotlib windows come with this functionality built-in, it is not
necessarily intuitive and lacks, e.g., support of the mouse wheel for
zooming. Hence, the plot widget implemented here aims at bridging this gap.


How to use the Plot widget in own GUIs?
=======================================

A rather minimal (and not very useful, though educational) example,
including the :class:`FileBrowser` as only widget of a main window:

.. code-block::

    import sys

    from PySide6 import QtWidgets

    import plot


    class MainWindow(QtWidgets.QMainWindow):

        def __init__(self):
            super().__init__()

            widget = plot.Plot()
            self.setCentralWidget(widget)

            import numpy as np

            t = np.linspace(0, 4*np.pi, 501)
            widget.axes.plot(t, np.sin(t), ".")

            self.show()


    if __name__ == "__main__":
        app = QtWidgets.QApplication(sys.argv)
        w = MainWindow()
        app.exec()


Of course, an import statement buried in a class is nothing you should
ever do in productive code. Nevertheless, it shows in minimalistic fashion
how to use the :class:`Plot` widget. As you can see, the figure and axes
properties are exposed as :attr:`Plot.figure` and :attr:`Plot.axes`,
and as these are the respective Matplotlib objects, they behave exactly as
expected.


Notes for developers
====================

TBD

.. todo::
    Implement mouse wheel for zooming (& panning) in Matplotlib canvas.

    https://stackoverflow.com/questions/11551049/


Module documentation
====================

"""
import collections.abc
import sys

import numpy as np

# Import PySide6 before matplotlib
from PySide6 import QtWidgets

from matplotlib import figure, widgets
from matplotlib.backends import backend_qtagg as backend

from qtbricks import utils


class Plot(QtWidgets.QWidget):
    """
    Plot widget for including Matplotlib plots in a GUI.

    Besides, the actual figure canvas containing the figure, a series of
    buttons is attached to the top, similar to the default toolbar of a
    Matplotlib window.

    As the widget exposes two public attributes, :attr:`figure` and
    :attr:`axes`, plotting should be fairly straight-forward and basically
    identical to other plots.

    If you need to change the axes, *e.g.* for multiple axes in a figure,
    this can be achieved by simply clearing the figure and creating new axes
    via the :meth:`matplotlib.figure.Figure.subplots` method:

    .. code-block::

        widget = Plot()
        widget.figure.clf()
        axes = widget.figure.subplots(2, 1)

        axes[0].plot(t, np.sin(t), ".")
        axes[1].plot(t, np.cos(t), ".")

    """

    def __init__(self):
        super().__init__()
        self._canvas = _FigureCanvas()

        self._setup_ui(self._canvas)

    @property
    def figure(self):
        """
        Matplotlib figure containing the actual plot.

        You can use this property to both, query and set all properties of
        the corresponding figure.

        Returns
        -------
        figure : :class:`matplotlib.figure.Figure`

        """
        return self._canvas.figure

    @property
    def axes(self):
        """
        Matplotlib axes for plotting data.

        You can use this axes to directly plot to them, as you can do with
        any other Matplotlib axes.

        Furthermore, you can (re)set these axes, in case you need to have
        several axes in one figure. In this case, the axes will be set for
        the underlying :class:`_FigureCanvas` object.

        Returns
        -------
        axes : :class:`matplotlib.axes.Axes`

        """
        return self._canvas.axes

    @axes.setter
    def axes(self, axes):
        self._canvas.axes = axes

    def refresh(self, force=False):
        """
        Refresh figure canvas.

        After each change in a plot, the canvas needs to be updated. While
        the matplotlib.pyplot interface takes care of this, using matplotlib
        from within a GUI sometimes requires to take care of this yourself.

        Parameters
        ----------
        force : :class:`bool`
            Whether to force a new draw of the canvas.

            There are two ways to refresh a canvas: force an immediate draw
            using :meth:`matplotlib.backend_bases.FigureCanvasBase.draw` and
            a more gentle draw once control returns to the GUI event loop
            using :meth:`matplotlib.backend_bases.FigureCanvasBase.draw_idle`.

            Default: False

        """
        if force:
            self._canvas.draw()
        else:
            self._canvas.draw_idle()

    def _setup_ui(self, figure_canvas):
        mpl_toolbar = backend.NavigationToolbar2QT(figure_canvas, None)
        home_button = utils.create_button(
            icon="house.svg",
            shortcut="h",
            tooltip="Reset plot to default settings",
            slot=mpl_toolbar.home,
        )
        back_button = utils.create_button(
            icon="backward-step.svg",
            tooltip="Undo last change",
            slot=mpl_toolbar.back,
        )
        forward_button = utils.create_button(
            icon="forward-step.svg",
            tooltip="Redo last change",
            slot=mpl_toolbar.forward,
        )
        # Button group
        pan_button = utils.create_button(
            icon="arrows-up-down-left-right.svg",
            shortcut="p",
            tooltip="Pan plot",
            slot=mpl_toolbar.pan,
            checkable=True,
        )
        zoom_button = utils.create_button(
            icon="magnifying-glass.svg",
            shortcut="z",
            tooltip="Zoom plot",
            slot=mpl_toolbar.zoom,
            checkable=True,
        )
        crosshair_button = utils.create_button(
            icon="plus.svg",
            shortcut="x",
            tooltip="Show crosshair cursor",
            slot=self._canvas.toggle_crosshair_cursor,
            checkable=True,
        )
        toggle_group = QtWidgets.QButtonGroup(self)
        toggle_group.addButton(pan_button)
        toggle_group.addButton(zoom_button)
        toggle_group.addButton(crosshair_button)
        utils.make_buttons_in_group_uncheckable(toggle_group)
        # /Button group
        grid_button = utils.create_button(
            icon="hashtag.svg",
            shortcut="#",
            tooltip="Show grid",
            slot=self._canvas.toggle_grid,
            checkable=True,
        )
        subplots_button = utils.create_button(
            icon="chart-line.svg",
            tooltip="Configure subplots",
            slot=mpl_toolbar.configure_subplots,
        )
        customise_button = utils.create_button(
            icon="sliders.svg",
            tooltip="Customise plot appearance",
            slot=mpl_toolbar.edit_parameters,
        )
        save_button = utils.create_button(
            icon="floppy-disk.svg",
            tooltip="Save plot",
            slot=mpl_toolbar.save_figure,
        )

        controls_layout = QtWidgets.QHBoxLayout()
        controls_layout.addWidget(home_button)
        controls_layout.addWidget(back_button)
        controls_layout.addWidget(forward_button)
        controls_layout.addWidget(zoom_button)
        controls_layout.addWidget(pan_button)
        controls_layout.addWidget(crosshair_button)
        controls_layout.addWidget(grid_button)
        controls_layout.addWidget(subplots_button)
        controls_layout.addWidget(customise_button)
        controls_layout.addWidget(save_button)
        controls_layout.addStretch()
        layout = QtWidgets.QVBoxLayout()
        layout.addLayout(controls_layout)
        layout.addWidget(figure_canvas)
        self.setLayout(layout)


class _FigureCanvas(backend.FigureCanvasQTAgg):
    """
    Figure canvas containing the Matplotlib figure for use within Qt GUIs.

    Attributes
    ----------
    figure : :class:`matplotlib.figure.Figure`
        Matplotlib figure containing the actual plot

    axes : :class:`matplotlib.axes.Axes`
        Matplotlib axes for plotting data

    """

    def __init__(self):
        self.figure = figure.Figure()  # figsize=(width, height), dpi=dpi
        self.axes = self.figure.add_subplot(111)
        super().__init__(self.figure)

        self._cursor = None
        self._background = None
        self._grid = None

    def toggle_crosshair_cursor(self):
        if self._cursor:
            for hline in self._cursor.hlines:
                hline.remove()
            for vline in self._cursor.vlines:
                vline.remove()
            self._cursor = None
            # noinspection PyUnresolvedReferences
            self.restore_region(self._background)
            self.blit(self.figure.bbox)
        else:
            # noinspection PyUnresolvedReferences
            self._background = self.figure.canvas.copy_from_bbox(
                self.figure.bbox
            )
            if isinstance(self.axes, collections.abc.Iterable):
                axes = self.axes
            else:
                axes = [self.axes]
            self._cursor = widgets.MultiCursor(
                canvas=None,
                axes=axes,
                useblit=False,
                horizOn=True,
                vertOn=True,
                color="red",
                linewidth=1,
            )

    def toggle_grid(self):
        if isinstance(self.axes, collections.abc.Iterable):
            for axes in self.axes:
                axes.grid(visible=self._grid)
        else:
            self.axes.grid(visible=self._grid)
        self.draw_idle()


class _MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()

        widget = Plot()
        self.setCentralWidget(widget)

        time = np.linspace(0, 4 * np.pi, 501)
        widget.axes.plot(time, np.sin(time), ".")

        widget.figure.clf()

        widget.axes = widget.figure.subplots(2, 1)
        widget.axes[0].plot(time, np.sin(time), ".")

        self.show()


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    w = _MainWindow()
    app.exec()



import unittest

from PySide6 import QtCore, QtWidgets

from qtbricks import testing


class TestQtestEnterText(unittest.TestCase):
    def setUp(self):
        self.app = (
            QtWidgets.QApplication.instance() or QtWidgets.QApplication()
        )
        self.widget = QtWidgets.QLineEdit()
        self.addCleanup(self.release_qt_resources)

    def release_qt_resources(self):
        self.widget.deleteLater()
        self.app.sendPostedEvents(event_type=QtCore.QEvent.DeferredDelete)
        self.app.processEvents()

    def test_qtest_enter_text_enters_text(self):
        text = "Lorem ipsum"
        testing.qtest_enter_text(widget=self.widget, text=text)
        self.assertEqual(text, self.widget.text())

    def test_qtest_enter_text_clears_text_before(self):
        text = "Lorem ipsum"
        self.widget.setText("Bla")
        testing.qtest_enter_text(widget=self.widget, text=text)
        self.assertEqual(text, self.widget.text())

    def test_qtest_enter_text_sends_signals(self):
        def mock_slot():
            self.has_been_called = True

        self.has_been_called = False
        text = "Lorem ipsum"
        self.widget.editingFinished.connect(mock_slot)
        testing.qtest_enter_text(widget=self.widget, text=text)
        self.assertTrue(self.has_been_called)



"""
File browser widget for selecting (multiple) files from a directory tree.

Sometimes we need a convenient file browser widget displaying a directory as
tree and allowing to both, navigate through the directory hierarchy and to
select (multiple) files. All we are usually interested in is the (full) file
names of the selected files.

There is currently just one public class in this module meant to be included
as a widget into your own GUIs: :class:`FileBrowser`. All other classes are
meant for internal use only, although they are documented.


General characteristics and intended purpose
============================================

The file browser widget is intended to be used as an interface to select one
or multiple individual *files* (not directories) and to return the filenames
with their full path, to further operate on this information.


How to use the FileBrowser in own GUIs?
=======================================

A rather minimal (and not very useful, though educational) example,
including the :class:`FileBrowser` as only widget of a main window:

.. code-block::

    import sys

    from PySide6 import QtWidgets

    from qtbricks import filebrowser


    class MainWindow(QtWidgets.QMainWindow):

        def __init__(self):
            super().__init__()

            widget = filebrowser.FileBrowser()

            # Purely for debugging purposes:
            widget.selection_changed.connect(lambda x: print("Selection:", x))
            self.setCentralWidget(widget)

            self.show()


    if __name__ == "__main__":
        app = QtWidgets.QApplication(sys.argv)
        w = MainWindow()
        app.exec()


While the :class:`FileBrowser` class emits
:attr:`FileBrowser.selection_changed` signals whenever the selection of its
items changed, you can always get the current selection from the
:attr:`FileBrowser.selection` attribute. The :attr:`FileBrowser.selection`
attribute contains the filenames including the full path for each selected
file (item). Similarly, the :attr:`FileBrowser.selection_changed` signal
contains the contents of :attr:`FileBrowser.selection` as payload.


Notes for developers
====================

Wherever possible, the naming follows :pep:`8` conventions. This naturally
leads to some inconsistencies between the PySide6 elements and the own code.

All icons used are from the FontAwesome font (version 6), with a CC BY 4.0
License (https://creativecommons.org/licenses/by/4.0/) for the icons
themselves. To decouple the location of the icons from the actual code,
the :func:`qtbricks.utils.image_path` function gets used.

The entire widget is laid out programmatically, *not* using QtDesigner.
Currently, all widgets are non-public instance attributes, allowing to
define them all together in the class constructor and factoring further
settings into separate methods.

Communication with the outside world beyond the widget primarily takes place
via the Qt signal--slot mechanism.

.. todo::
    Try to reimplement buttons and edit as a Qt toolbar with respective
    actions that can even be added to an external window?

.. todo::
    Try to reimplement toolbar/buttons&edit such that the edit moves in a
    second line if the widget gets too narrow.

.. todo::
    Provide (sensible) minimumSizeHint for FileBrowser widget.

    Depending on the context the widget is used in, this seems not
    necessary... Hence, lower priority.


Module documentation
====================

"""

import os
import sys

from PySide6 import QtWidgets, QtCore, QtGui

from qtbricks import utils


class FileBrowser(QtWidgets.QWidget):
    """
    File browser widget for selecting (multiple) files from a directory tree.

    Sometimes we need a convenient file browser widget displaying a
    directory as tree and allowing to both, navigate through the directory
    hierarchy and to select (multiple) files. All we are usually
    interested in is the (full) file names of the selected files.

    At the core of the widget is a tree view (a :class:`_FileTree` object),
    but for the convenience of the user, a series of additional control
    widgets is added on top:

    * Buttons for home, back, up, and forward
    * A line edit displaying the current directory path

    The back and forward buttons are only active once the user navigated
    within the directory tree. The line edit displaying the current
    directory path can be used as an input as well and will always be in
    sync with the tree view. If the user enters a non-existing directory,
    the edit will simply be reverted to the last (valid) path.

    Multiple selections are allowed, but only files can be selected,
    not directories. You can select multiple files by pressing the "Ctrl"
    key while clicking. In the same way, you can deselect a given file.
    Pressing the "Shift" key when clicking selects a range of files.

    Selecting files using the keyboard follows established conventions as
    well. Moving with the arrow keys moves the selection. Holding the
    "Ctrl" key when navigating with the arrow keys allows to select an
    additional file pressing the "Space" key. Holding the "Shift" key when
    navigating with the arrow keys selects an entire range of files. You can
    even use a sequence of holding "Ctrl" and "Shift" keys for arbitrarily
    complex selection patterns. Just make sure to always keep pressing
    either "Ctrl" or "Shift" when navigating, as otherwise, the selection
    will be cleared and only the current item selected. Deselecting an item
    is possible as well, holding the "Ctrl" key and using the "Space" key to
    toggle selection of the current item.

    Double-clicking on a directory will change the root path to this
    directory.

    Currently, the class has no public methods and only two public
    attributes and a signal documented below.

    Parameters
    ----------
    path : :class:`str`
        Root path to be set for the file browser

    Attributes
    ----------
    root_path : :class:`str`
        Root path set currently for the file browser

    selection : :class:`list`
        Names of the currently selected files

        The names are the actual full paths to the file on the file system.

        Thanks to using a list, the names should always appear in the order
        they have been selected.

    """

    selection_changed = QtCore.Signal(set)
    """
    Signal emitted when the selection of items changed.

    The signal contains the selection as :class:`set` parameter.
    """

    def __init__(self, path=""):
        super().__init__()

        self.root_path = path or os.path.abspath(os.path.curdir)
        self.selection = []

        self._previous_path = ""
        self._next_path = ""
        self._model_settings = {}

        self._home_button = QtWidgets.QPushButton()
        self._back_button = QtWidgets.QPushButton()
        self._up_button = QtWidgets.QPushButton()
        self._forward_button = QtWidgets.QPushButton()
        self._curdir_edit = QtWidgets.QLineEdit()
        self._tree_view = _FileTree(root_path=self.root_path)

        self._setup_ui()
        self._update_ui()

    @property
    def model_settings(self):
        """
        Settings for the underlying QFileSystemModel.

        Often, when browsing the file system, we want to control (and
        restrict) what is displayed how. To this end, settings for the
        underlying model need to be set.

        The settings are contained in a :class:`dict` that supports the
        following fields:

        filters : :class:`list`
            A list of strings to be used for filtering the files.

            Typical use cases would be filters for file extensions, such as
            ``*.py`` or ``*.png``.

            The QFileSystemModel class only supports basic wildcard
            filtering, so you will need to use a ``QSortFilterProxyModel`` to
            get fully customisable filtering. Note however, that the latter
            is currently *not* implemented.

            See https://stackoverflow.com/questions/72587813 for inspiration.

        filter_disables : :class:`bool`
            Whether files filtered with the above filter are displayed.

            The standard handling of the QFileSystemModel is to only
            disable, but not hide the filtered out entries. Set to
            ``False`` in case you want to *hide* the entries entirely.

        Returns
        -------
        model_settings : :class:`dict`
            Settings for the underlying QFileSystemModel


        Note
        ----

        The reason for this attribute to be a property is simple,
        yet deserves a comment for developers: You need to be able to
        set/alter the model settings after you have instantiated the
        class. Therefore, upon changing the settings, an internal method
        is called taking care of applying your settings to the model.

        """
        return self._model_settings

    @model_settings.setter
    def model_settings(self, settings):
        self._model_settings = settings
        self._tree_view.apply_settings(self._model_settings)

    def _setup_ui(self):
        self._set_widget_properties()
        self._set_layout()
        self._connect_signals()

    def _set_widget_properties(self):
        self._home_button.setIcon(QtGui.QIcon(utils.image_path("house.svg")))
        self._home_button.setToolTip(
            "Go to the home directory of the current user"
        )
        self._up_button.setIcon(
            QtGui.QIcon(utils.image_path("circle-up.svg"))
        )
        self._up_button.setToolTip("Go one directory up in the hierarchy")
        self._back_button.setIcon(
            QtGui.QIcon(utils.image_path("circle-left.svg"))
        )
        self._back_button.setToolTip("Go back to the previous directory")
        self._forward_button.setIcon(
            QtGui.QIcon(utils.image_path("circle-right.svg"))
        )
        self._forward_button.setToolTip(
            "Revert going back to the previous directory"
        )
        self._curdir_edit.setText(self.root_path)
        self._curdir_edit.setCursorPosition(len(self._curdir_edit.text()))
        self._curdir_edit.setToolTip(
            "Display/edit the current (root) directory.\n"
            "Only existing directories will be accepted as user input."
        )

    def _set_layout(self):
        controls_layout = QtWidgets.QHBoxLayout()
        controls_layout.addWidget(self._home_button)
        controls_layout.addWidget(self._back_button)
        controls_layout.addWidget(self._up_button)
        controls_layout.addWidget(self._forward_button)
        controls_layout.addStretch()
        layout = QtWidgets.QVBoxLayout()
        layout.addLayout(controls_layout)
        layout.addWidget(self._curdir_edit)
        layout.addWidget(self._tree_view)
        self.setLayout(layout)

    def _connect_signals(self):
        self._up_button.pressed.connect(self._move_up)
        self._home_button.pressed.connect(self._go_home)
        self._back_button.pressed.connect(self._go_back)
        self._forward_button.pressed.connect(self._go_forward)
        self._curdir_edit.editingFinished.connect(self._change_path)
        self._tree_view.root_path_changed.connect(self._change_root_path)
        self._tree_view.selection_changed.connect(self._change_selection)

    def _change_path(self):
        new_path = self._curdir_edit.text()
        if os.path.exists(new_path):
            self._change_root_path(new_path)
        else:
            self._curdir_edit.setText(self.root_path)

    def _change_selection(self, selection: list):
        self.selection_changed.emit(selection)
        self.selection = selection

    def _change_root_path(self, path=""):
        if path == self.root_path:
            return
        if path.endswith("/") and path != "/":
            path = path[:-1]
        self._previous_path = self.root_path
        self.root_path = path
        self._update_ui()

    def _move_up(self):
        self._change_root_path(path=os.path.split(self.root_path)[:-1][0])

    def _go_home(self):
        self._change_root_path(path=os.path.expanduser("~"))

    def _go_back(self):
        self._next_path = self.root_path
        self._change_root_path(path=self._previous_path)

    def _go_forward(self):
        path = self._next_path
        self._next_path = ""
        self._change_root_path(path=path)

    def _update_ui(self):
        self._curdir_edit.setText(self.root_path)
        self._tree_view.set_root_path(self.root_path)
        self._back_button.setEnabled(bool(self._previous_path))
        self._forward_button.setEnabled(bool(self._next_path))


class _FileTree(QtWidgets.QTreeView):
    """
    Tree view of the file system emitting names of selected files.

    The tree view is based on :class:`PySide6.QtWidgets.QTreeView` and uses
    :class:`PySide6.QtWidgets.QFileSystemModel` as underlying model for the
    file system. Furthermore, multiple selections are allowed, but only
    files can be selected, not directories.

    As a selection mode,
    :attr:`QtWidgets.QAbstractItemView.ExtendedSelection` is used, meaning
    that you can select multiple files by pressing the "Ctrl" key while
    clicking. In the same way, you can deselect a given file. Pressing the
    "Shift" key when clicking selects a range of files.

    Double-clicking on a directory will change the root path to this
    directory.

    The class emits two signals described below.

    Parameters
    ----------
    root_path : :class:`str`
        Root path to be set for the file browser


    .. todo::
        Properly handle hiding of columns, using :meth:`self.hideColumn(#)`.

    .. todo::
        Context menu, allowing to set the columns to be displayed?

    .. todo::
        Allow using the "Return" key to enter directories?

    """

    root_path_changed = QtCore.Signal(str)
    """
    Signal emitted when the root path of the underlying model changed.

    The signal contains the new root path as :class:`str` parameter.
    """

    selection_changed = QtCore.Signal(list)
    """
    Signal emitted when the selection of items changed.

    The signal contains the selection as :class:`list` parameter.
    """

    def __init__(self, root_path=""):
        super().__init__()
        self._root_path = root_path
        self._model = _FileSystemModel()
        self._setup_ui()

    def _setup_ui(self):
        self._model.rootPathChanged.connect(self._root_path_changed)
        self.setModel(self._model)

        self.set_root_path(self._root_path)
        # self.hideColumn(1)

        self.setExpandsOnDoubleClick(False)
        # noinspection PyUnresolvedReferences
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.doubleClicked.connect(self._double_clicked)

        self.setAllColumnsShowFocus(True)

    def set_root_path(self, path=""):
        """
        Set the root path of the directory tree view.

        Parameters
        ----------
        path : :class:`str`
            Path to be set as root path.

        """
        if not path:
            return
        self._model.setRootPath(path)
        self.setRootIndex(self._model.index(self._model.rootPath()))

    def _double_clicked(self, index=QtCore.QModelIndex()):
        """
        Handle double-click events on items.

        In case the user double-clicks on a directory, the root path of the
        tree view is reset to this directory.

        Note that due to the tree view, the user doesn't need to
        double-click on a directory to see its contents, but can unfold it
        as well.

        Parameters
        ----------
        index : :class:`QtCore.QModelIndex`
            Index of the selected item

        """
        if self._model.fileInfo(index).isDir():
            self.set_root_path(self._model.filePath(index))

    def _root_path_changed(self, path=""):
        """
        Handle "rootPathChanged" signal of the underlying model.

        Whenever the root path of the model changed,
        a :attr:`root_path_changed` signal will be emitted sending the new
        path as payload.

        The signal will only be emitted if the path actually changed.

        Parameters
        ----------
        path : :class:`str`
            New root path that has been set.

        """
        if path != self._root_path:
            self._root_path = path
            self.root_path_changed.emit(path)

    def selectionCommand(  # noqa N802
        self, index=QtCore.QModelIndex, event=QtCore.QEvent
    ):
        """
        Handle selection of the underlying tree view.

        In case a directory has been selected, it will automatically be
        deselected, thus making directories not selectable. The idea behind
        this behaviour is to return only files, not directories,
        for settings where you want to operate on one or multiple files,
        but not on entire directories.

        For all items that are not directories, the super method is called
        and the parameters passed to this method.

        Parameters
        ----------
        index : :class:`QtCore.QModelIndex`
            Index of the selected item

        event : :class:`QtCore.Event`
            Event sent during selection.

        """
        if self._model.fileInfo(index).isDir():
            # noinspection PyUnresolvedReferences
            return QtCore.QItemSelectionModel.Deselect
        return super().selectionCommand(index, event)

    def selectionChanged(self, selected, deselected):  # noqa N802
        """
        Handle changes of selection of the underlying tree view.

        A :attr:`selection_changed` signal is emitted, conveying the list of
        filenames (with their full path) corresponding to the currently
        selected items.

        Afterwards, the super method is called and the parameters passed to
        this method.

        Parameters
        ----------
        selected : :class:`QtCore.QItemSelection`
            Selected item

        deselected: :class:`QtCore.QItemSelection`
            Deselected item

        """
        selected_rows = []
        for index in self.selectedIndexes():
            item = self._model.filePath(index)
            if item not in selected_rows:
                selected_rows.append(item)
        self.selection_changed.emit(selected_rows)
        super().selectionChanged(selected, deselected)

    def apply_settings(self, model_settings):
        if "filters" in model_settings:
            self._model.setNameFilters(model_settings["filters"])
        if "filter_disables" in model_settings:
            self._model.setNameFilterDisables(
                model_settings["filter_disables"]
            )


class _FileSystemModel(QtWidgets.QFileSystemModel):
    """
    Model of the file system used in the tree view.

    Basically, the class is identical to its base class, currently just
    adding the relevant code for displaying tooltips.
    """

    # noinspection PyUnresolvedReferences,PyMethodOverriding
    def data(self, index, role):
        """
        Text or else used to display the selected item in the given context.

        Parameters
        ----------
        index : :class:`QtCore.QModelIndex`
            Index of the selected item

        role : :class:`int`
            The Qt role deciding about the context

        Returns
        -------
        data : :class:`Any`
            Data used to display the item

        """
        if role == QtCore.Qt.ToolTipRole:
            role = QtCore.Qt.DisplayRole
        return super().data(index, role)


class _MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()

        widget = FileBrowser(os.path.dirname(__file__))
        # Purely for debugging purposes:
        widget.selection_changed.connect(lambda x: print("Selection:", x))
        self.setCentralWidget(widget)

        self.show()


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    w = _MainWindow()
    app.exec()



"""
Small Qt widgets for general use with Qt GUIs.

There is a number of small widgets that can be created with a few lines of code
from existing Qt widgets. However, it is often convenient to have these widgets
as separate classes.

Typical examples include horizontal and vertical lines that are basically
one-dimensional instances of QFrame.


Widgets
=======

Currently, the following widgets are contained in this module:

* :class:`QHLine`

    Horizontal line for Qt GUIs

* :class:`QVLine`

    Vertical line for Qt GUIs


Module documentation
====================

"""

from PySide6 import QtWidgets


class QHLine(QtWidgets.QFrame):
    """
    Horizontal line for Qt GUIs.

    Sometimes there is the need to add a horizontal line to a Qt layout.
    As a horizontal line is nothing else than a QFrame object, this class
    provides a convenient shortcut.

    Inspiration taken from Michael Leonard via StackOverflow:
    https://stackoverflow.com/a/41068447

    Examples
    --------
    Suppose you want to add a horizontal line to your grid layout

    .. code-block::

        from PySide6 import QtWidgets
        import qtbricks


        layout = QtWidgets.QGridLayout()
        layout.addWidget(qtbricks.widgets.QHLine(), 4, 0, 1, 2)


    Here, the horizontal line will span two columns of your grid layout.


    See Also
    --------
    :class:`QVLine`
        Vertical line for Qt GUIs

    """

    def __init__(self):
        super().__init__()
        self.setFrameShape(QtWidgets.QFrame.HLine)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)


class QVLine(QtWidgets.QFrame):
    """
    Vertical line for Qt GUIs.

    Sometimes there is the need to add a vertical line to a Qt layout.
    As a horizontal line is nothing else than a QFrame object, this class
    provides a convenient shortcut.

    Inspiration taken from Michael Leonard via StackOverflow:
    https://stackoverflow.com/a/41068447


    Examples
    --------
    Suppose you want to add a vertical line to your grid layout

    .. code-block::

        from PySide6 import QtWidgets
        import qtbricks


        layout = QtWidgets.QGridLayout()
        layout.addWidget(qtbricks.widgets.QVLine(), 4, 0, 2, 1)


    Here, the vertical line will span two rows of your grid layout.


    See Also
    --------
    :class:`QHLine`
        Horizontal line for Qt GUIs

    """

    def __init__(self):
        super().__init__()
        self.setFrameShape(QtWidgets.QFrame.VLine)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)



import os
import setuptools


def read(filename):
    with open(os.path.join(os.path.dirname(__file__), filename)) as file:
        content = file.read()
    return content


setuptools.setup(
    name="qtbricks",
    version=read("VERSION").strip(),
    description="Python Qt components: widgets and utils, focussing on "
    "scientific GUIs",
    long_description=read("README.rst"),
    long_description_content_type="text/x-rst",
    author="Till Biskup",
    author_email="code@till-biskup.de",
    url="https://pypi.org/project/qtbricks/",
    project_urls={
        "Documentation": "https://qtbricks.docs.till-biskup.de/",
        "Source": "https://github.com/tillbiskup/qtbricks",
    },
    packages=setuptools.find_packages(exclude=("tests", "docs")),
    license="BSD",
    keywords=[
        "Qt",
        "widgets",
        "building blocks",
        "PySide6",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development",
        "Development Status :: 4 - Beta",
    ],
    install_requires=[
        "PySide6",
        "matplotlib",
    ],
    extras_require={
        "dev": [
            "prospector",
            "pyroma",
            "bandit",
            "black",
            "pymetacode",
        ],
        "docs": [
            "sphinx",
            "sphinx-rtd-theme",
            "sphinx_multiversion",
        ],
        "deployment": [
            "build",
            "twine",
        ],
    },
    python_requires=">=3.9",
    include_package_data=True,
)



import unittest

from PySide6 import QtCore, QtWidgets

from qtbricks import mainwindow


class TestMainWindow(unittest.TestCase):
    def setUp(self):
        self.app = (
            QtWidgets.QApplication.instance() or QtWidgets.QApplication()
        )
        self.window = mainwindow.MainWindow()
        self.addCleanup(self.release_qt_resources)

    def release_qt_resources(self):
        self.window.deleteLater()
        self.app.sendPostedEvents(event_type=QtCore.QEvent.DeferredDelete)
        self.app.processEvents()

    def test_instantiate_class(self):
        pass


class TestGeneralDockWindow(unittest.TestCase):
    def setUp(self):
        self.app = (
            QtWidgets.QApplication.instance() or QtWidgets.QApplication()
        )
        self.window = mainwindow.GeneralDockWindow()
        self.addCleanup(self.release_qt_resources)

    def release_qt_resources(self):
        self.window.deleteLater()
        self.app.sendPostedEvents(event_type=QtCore.QEvent.DeferredDelete)
        self.app.processEvents()

    def test_instantiate_class(self):
        pass

    def test_initialisation_sets_window_title(self):
        window_title = "Lorem ipsum"
        window = mainwindow.GeneralDockWindow(title="Lorem ipsum")
        self.assertEqual(window.windowTitle(), window_title)

    def test_initialisation_sets_widget(self):
        widget = QtWidgets.QListWidget()
        window = mainwindow.GeneralDockWindow(widget=widget)
        self.assertEqual(window.widget(), widget)

    def test_initialisation_sets_object_name(self):
        object_name = "Foo"
        window = mainwindow.GeneralDockWindow(object_name=object_name)
        self.assertEqual(window.objectName(), object_name)



"""
Help-About dialog for main GUI windows.

**Purpose:** Typically, every GUI has a "Help About" dialog displaying some
information regarding the application, such as its name and version number,
website, license, authors, and debug information.

**Design principles:** The dialog should be as self-contained and
self-consistent as possible. The only crucial information that needs to be
set from the outside is the package name.

**Limitations:** Handling of author/contributor names and email addresses
is limited by how this information is stored in the ``setup.py`` file and
read by the ``importlib.metadata`` module.


Some notes for developers
=========================

If you base the main window of your GUI application on the
:class:`qtbricks.mainwindow.MainWindow` class and set the
:attr:`qtbricks.mainwindow.MainWindow.package_name` attribute accordingly,
you are basically set up.

If, however, you want to change or extend the texts displayed in the three
tabs, the easiest way is to change each of the respective methods returning
the text to be displayed, namely:

* :meth:`AboutDialog._create_about_text`
* :meth:`AboutDialog._create_authors_text`
* :meth:`AboutDialog._create_debug_info_text`

Each of these non-public methods returns a text that is set as text of the
underlying display widget. Hence, you can use a reduced set of HTML tags to
properly format the text.

Furthermore, if you need more control, you may think of using `Jinja
<https://jinja.palletsprojects.com/>`_ as template engine.


Module documentation
====================

"""
from importlib import metadata
import platform
import string
import subprocess  # noqa
import sys

from PySide6 import QtWidgets, QtGui

import qtbricks.utils

ABOUT_STRING = """
<p>
${package_name} - Version ${package_version}
</p>

<p>
<em>${description}</em>
</p>

<p>
Website: <a href="${website}" title="Open ${website} in your preferred 
browser (if configured).">${website}</a>
</p>

<p>
License: ${package_name} is free software: you can redistribute it and/or 
modify it under the terms of the <strong>${license} license</strong>.
</p>
"""

AUTHOR_STRING = """
<p>
The following people contributed to ${package_name}:
</p>

${authors}

You may contact the authors or maintainer(s) using the following email 
address: <a href="mailto:${email}" title="Send email to ${email} 
using your preferred email client (if configured).">${email}</a>.

<hr />

<p>
License: ${package_name} is free software: you can redistribute it and/or 
modify it under the terms of the <strong>${license} license</strong>.
</p>
"""

DEBUG_INFO_STRING = """
<p>
${package_name} - Version ${package_version}
</p>

<p>
OS: ${os}<br />
CPU architecture: ${architecture}<br />
Kernel: ${kernel}
</p>

<p>
Python version: ${python_version}<br />
PySide6 version: ${pyside6_version}
</p>

<hr />

<p>
All installed Python packages:
</p>

<p>
${all_packages}
</p>
"""


class AboutDialog(QtWidgets.QDialog):
    """
    Typical "Help About" Dialog of the main window of a GUI application.

    Most GUI main windows have a "Help About" dialog displaying some
    information regarding the application, such as its name and version
    number, website, license, authors, and debug information.

    Attributes
    ----------
    package_name : :class:`str`
        Name of the package

        This information is crucial for displaying the relevant
        package metadata, such as version, short description, authors,
        and licence.

    logo : :class:`str`
        Path to logo image file


    Parameters
    ----------
    parent : :class:`PySide6.QtWidgets.QWidget`
        Parent of the dialog

        The dialog will usually be centered upon the parent.

    package_name : :class:`str`
        Name of the package

        This information is crucial for displaying the relevant
        package metadata, such as version, short description, authors,
        and licence.

    logo : :class:`str`
        Path to logo image file

    """

    def __init__(self, parent=None, package_name="", logo=""):
        super().__init__(parent)

        self.package_name = package_name
        if logo:
            self.logo = QtGui.QPixmap(logo)
        else:
            self.logo = QtGui.QPixmap(qtbricks.utils.image_path("icon.svg"))

        self._application_name = QtWidgets.QApplication.applicationName()
        self._dialog_buttons = QtWidgets.QDialogButtonBox.Close
        self._button_box = QtWidgets.QDialogButtonBox(self._dialog_buttons)

        self._top_layout = QtWidgets.QHBoxLayout()
        self._tab_widget = QtWidgets.QTabWidget()
        self._tab_about = QtWidgets.QTextBrowser()
        self._tab_authors = QtWidgets.QTextBrowser()
        self._tab_debug_info = QtWidgets.QTextEdit()

        self._setup_ui()

    def _setup_ui(self):
        """
        Setup the dialog window.

        This method takes care of setting up all the elements of the dialog.
        This is a three-step process, each carried out calling the
        corresponding non-public method:

        #. Set the widget properties
        #. Set the layout
        #. Connect the signals and slots

        A requirement is to define all widgets as non-public attributes in
        the class constructor. This comes with the advantage to separate
        the different tasks into methods.
        """
        self._create_top_layout()
        self._add_tab_widgets()
        self._set_widget_properties()
        self._set_layout()
        self._connect_signals()

    def _create_top_layout(self):
        package_metadata = metadata.metadata(self.package_name)
        version = package_metadata["Version"]
        icon = QtWidgets.QLabel()
        icon.setPixmap(self.logo.scaledToHeight(60))
        text = QtWidgets.QLabel(
            f"<h1>{self.package_name}</h1>Version {version}"
        )
        self._top_layout.addWidget(icon)
        self._top_layout.addWidget(text)
        self._top_layout.addStretch(1)

    def _add_tab_widgets(self):
        self._tab_widget.addTab(self._tab_about, "About")
        self._tab_widget.addTab(self._tab_authors, "Authors")
        self._tab_widget.addTab(self._tab_debug_info, "Debug info")

    def _set_widget_properties(self):
        """
        Set the widgets of all the UI components.

        Usually, a widget will contain a number of other widgets whose
        properties need to be set initially. This is the one central place
        to do this.

        A requirement is to define all widgets as non-public attributes in
        the class constructor.
        """
        self.setWindowTitle(f"About {self.package_name}")
        self._tab_about.setOpenExternalLinks(True)
        self._tab_about.setReadOnly(True)
        self._tab_about.setFrameShape(QtWidgets.QFrame.NoFrame)
        self._tab_about.setText(self._create_about_text())
        self._tab_authors.setOpenExternalLinks(True)
        self._tab_authors.setReadOnly(True)
        self._tab_authors.setFrameShape(QtWidgets.QFrame.NoFrame)
        self._tab_authors.setText(self._create_authors_text())
        self._tab_debug_info.setReadOnly(True)
        self._tab_debug_info.setFrameShape(QtWidgets.QFrame.NoFrame)
        self._tab_debug_info.setText(self._create_debug_info_text())

    def _set_layout(self):
        """
        Lay out the elements of the dialog.

        Usually, a dialog will contain a number of other widgets that need
        to be laid out in some way. This is the central place to do this.

        Furthermore, typically a dialog has a series of buttons contained
        in a button box that need to be placed somewhere as well.

        A requirement is to define all widgets as non-public attributes in
        the class constructor.
        """
        layout = QtWidgets.QVBoxLayout()
        layout.addLayout(self._top_layout)
        layout.addWidget(self._tab_widget)
        layout.addWidget(self._button_box)
        self.setLayout(layout)

    def _connect_signals(self):
        """
        Connect all signals and slots of the dialog.

        As a bare minimum, the signals of the individual buttons need to be
        connected to some slot. In case of a single "Close" button,
        there is a special (not necessarily intuitive) signal to be connected:

        .. code-block::

            self._button_box.rejected.connect(self.reject)

        A requirement is to define all widgets as non-public attributes in
        the class constructor.
        """
        self._button_box.rejected.connect(self.reject)

    def _create_about_text(self):
        package_metadata = metadata.metadata(self.package_name)
        text = string.Template(ABOUT_STRING)
        substitutions = {
            "package_name": self.package_name,
            "package_version": package_metadata["Version"],
            "description": package_metadata["Summary"] or "Lorem ipsum...",
            "website": package_metadata["Home-page"] or "http://example.org/",
            "license": package_metadata["License"],
        }
        return text.substitute(substitutions)

    def _create_authors_text(self):
        package_metadata = metadata.metadata(self.package_name)
        text = string.Template(AUTHOR_STRING)
        author_names = "<li>John Doe</li>\n"
        if "Author" in package_metadata:
            author_names = [
                f"<li>{name}</li>\n"
                for name in package_metadata["Author"].split(",")
            ]
        authors = f"<ul>\n{author_names[0]}</ul>\n"
        if "Author-email" in package_metadata:
            email = package_metadata["Author-email"]
        else:
            email = "doe@example.net"
        substitutions = {
            "package_name": self.package_name,
            "authors": authors,
            "email": email,
            "license": package_metadata["License"],
        }
        return text.substitute(substitutions)

    def _create_debug_info_text(self):
        package_metadata = metadata.metadata(self.package_name)
        text = string.Template(DEBUG_INFO_STRING)
        substitutions = {
            "package_name": self.package_name,
            "package_version": package_metadata["Version"],
            "os": platform.system(),
            "architecture": platform.machine(),
            "kernel": platform.release(),
            "python_version": platform.python_version(),
            "pyside6_version": metadata.metadata("PySide6")["Version"],
            "all_packages": self._package_list(),
        }
        return text.substitute(substitutions)

    @staticmethod
    def _package_list():
        """
        Return list of all packages currently installed with version number.

        There are several ways to get the result of "pip list"
        programmatically from within Python. However, getting the version
        numbers as well seems to require using the deprecated
        ``pkg_resources``, as ``pkg_resources.working_set`` seems to have
        no replacement in other modules. Hence, we rely here on calling
        "pip list" directly from the command line and catching the output.

        The output is reformatted for use in an HTML context. Hence,
        the method in its present form is not a general-purpose function.

        Returns
        -------
        package_list : :class:`str`
            List of all packages currently installed

        """
        args = [sys.executable, "-m", "pip", "list"]
        process = subprocess.run(  # noqa
            args,
            check=True,
            capture_output=True,
        )
        output = process.stdout.decode().split("\n")[2:]
        package_list = "<br />\n".join(
            [": ".join(package.split()[0:2]) for package in output if output]
        )
        return package_list


if __name__ == "__main__":
    app = QtWidgets.QApplication()
    app.setOrganizationName("qtbricks")
    app.setOrganizationDomain("example.org")
    app.setApplicationName("Demo application")
    app.setWindowIcon(QtGui.QIcon(qtbricks.utils.image_path("icon.svg")))

    dialog = AboutDialog(package_name="qtbricks")
    dialog.exec()
    del app
